"""
This module brings a Python interface to the tool `"StableMotifs"
<https://github.com/jgtz/StableMotifs>`_ (Jorge G. T.  Zañudo and Réka Albert),
for the control of Boolean networks.

The control predictions can be processed using the `algorecell_types
<https://github.com/algorecell/algorecell_types>`_ library, which eases the
display and comparison with other control methods.

Installation instructions at https://github.com/algorecell/StableMotifs-python

Quick usage:

>>> import stablemotifs

Model loading:

>>> sm = stablemotifs.load("network.txt") # in BooleanNet format
# alternatively, load with biolqm in any format
>>> import biolqm
>>> lm = biolqm.load("model.zginml") # or any format support by bioLQM
>>> sm = stablemotifs.load(lm)

Reprogramming predictions:

>>> sr = sm.reprogramming_to_attractor({"A": 1, "B": 0})
>>> sr.as_table()

See ``help(sr)`` for other display methods

Examples can be found at:
    * https://nbviewer.jupyter.org/github/algorecell/StableMotifs-python/tree/master/examples/

"""

import atexit
import os
import shutil
import subprocess
import sys
import tempfile

from colomoto_jupyter import import_colomoto_tool
from colomoto_jupyter.io import ensure_localfile

from colomoto.minibn import BooleanNetwork

from .jupyter import upload

from .results import StableMotifsResult

def load(model, fixed=None, mcl="", msm="", quiet=False):
    """
    Execute StableMotifs analysis on the given Boolean network model.

    :param model: either a ``biolqm`` or ``ginsim`` object, or filename/URL of Boolean network in BooleanNet format
    :keyword dict[str,int] fixed: fix the given nodes to their associated given values
    :keyword str mcl: Optional threshold in the maximum cycle length (mcl). One must specify both a mcl and msm.
    :keyword str msm: Optional threshold in the maximum stable motif size (msm).  One must specify both a mcl and msm.
    :keyword bool quiet: If True, skip computation output
    :rtype: :py:class:`.results.StableMotifsResult` instance
    """

    # prepare temporary working space
    wd = tempfile.mkdtemp(prefix="StableMotifs-")
    # cleanup working directory at exit
    def cleanup():
        shutil.rmtree(wd)
    atexit.register(cleanup)

    def biolqm_import(biolqm, lqm):
        if fixed:
            pert = " ".join((f"{node}%{value}" for node, value in fixed.items()))
            lqm = biolqm.perturbation(lqm, pert)
        modelfile = os.path.join(wd, "model.txt")
        assert biolqm.save(lqm, modelfile, "booleannet"), "Error converting from bioLQM"
        return modelfile, False

    is_modelfile = True
    if isinstance(model, BooleanNetwork):
        lqm = model.to_biolqm()
        biolqm = import_colomoto_tool("biolqm")
        modelfile, is_modelfile = biolqm_import(biolqm, lqm)
    elif "biolqm" in sys.modules:
        biolqm = sys.modules["biolqm"]
        if biolqm.is_biolqm_object(model):
            modelfile, is_modelfile = biolqm_import(biolqm, model)
    if is_modelfile and "ginsim" in sys.modules:
        ginsim = sys.modules["ginsim"]
        if ginsim.is_ginsim_object(model):
            model = ginsim.to_biolqm(model)
            biolqm = import_colomoto_tool("biolqm")
            modelfile, is_modelfile = biolqm_import(biolqm, model)

    if is_modelfile:
        modelfile = ensure_localfile(model)
        if fixed:
            biolqm = import_colomoto_tool("biolqm")
            model = biolqm.load(modelfile, "booleannet")
            modelfile, is_modelfile = biolqm_import(biolqm, model)
        else:
            shutil.copy(modelfile, wd)

    modelbase = os.path.basename(modelfile)
    model_name = modelbase.split(".")[0]

    # invoke StableMotifs
    argv = ["StableMotifs", modelbase]
    if mcl and msm:
        argv += list(map(str, [mcl, msm]))
    proc = subprocess.Popen(argv, cwd=wd,
                stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    with proc.stdout:
        for line in iter(proc.stdout.readline, b''):
            if quiet:
                continue
            sys.stdout.write(line)
    assert proc.wait() == 0, "An error occured while running StableMotifs"

    # return interface to results
    return StableMotifsResult(wd, model_name, fixed)





import os

import pandas as pd

from algorecell_types import *
from colomoto.types import PartialState

class StableMotifsResult(object):
    """
    Class of objects returned by :py:func:`load` function giving access to
    StableMotifs results.

    .. py:attribute:: attractors

        List of attractors (dictionnaries mapping node names to ``0``, ``1``, or
        ``'X'``).
    """
    def __init__(self, wd, model_name, fixed):
        self.wd = wd
        self.model_name = model_name
        self.fixed = fixed
        self.attractors = self._parse_attractors()

    def _wfile(self, pattern):
        return os.path.join(self.wd, pattern%self.model_name)

    def _parse_attractors(self):
        inputf = self._wfile("%s-QuasiAttractors.txt")
        df = pd.read_csv(inputf, sep='\t', index_col=-1) #index_col=-1 we cheat by making the extra tabs at the end of the rows in the files as the index column, and then resetting it to just natural incrementing integers
        df.reset_index(inplace=True,drop=True)
        d = df.to_dict("index")
        return [d[i] for i in range(len(d))]

    def _parse_control_sets(self):
        inputf = self._wfile("%s-StableMotifControlSets.txt")
        control_sets = dict([(i, []) for i in range(len(self.attractors))])
        with open(inputf) as f:
            for c in f.readlines():
                if not c:
                    continue
                cc=c.split('\t')[0]
                cd=[(i.split('=')[0],int(i.split('=')[1])) for i in cc.split(' ')]
                attr=c.split('\t')[1]
                attr_id = int(attr[9:])
                control_sets[attr_id].append(dict(sorted(cd)))
        return control_sets

    def _parse_stable_motifs(self):
        inputf = self._wfile("%s-StableMotifs.txt")
        stable_motifs = []
        with open(inputf) as f:
            for sms in f.readlines():
                if not sms.strip():
                    continue
                smss=[(i.split('=')[0],int(i.split('=')[1])) for i in sms.strip().split('\t')]
                stable_motifs.append(dict(smss))
        return stable_motifs

    @property
    def stable_motifs(self):
        """
        List of stable motifs, mapping subset of nodes to Boolean values (`dict[str,int]`)
        """
        if not hasattr(self, "_StableMotifsResult__cache_stable_motifs"):
            self.__cache_stable_motifs = self._parse_stable_motifs()
        return self.__cache_stable_motifs

    @property
    def control_sets(self):
        """
        Dictionnary mapping attractor indices to their control sets (`dict[str,int] list`)
        """
        if not hasattr(self, "_StableMotifsResult__cache_control_sets"):
            self.__cache_control_sets = self._parse_control_sets()
        return self.__cache_control_sets

    def reprogramming_to_attractor(self, *spec, **kwspec):
        """
        Compute reprogramming strategies ensuring that only attractors matching
        with the given specification can be reached with the asynchronous mode.
        The perturbations are temporary, i.e., can be released once in the
        attractor.

        :rtype: `algorecell_types.ReprogrammingStrategies <https://algorecell-types.readthedocs.io/#algorecell_types.ReprogrammingStrategies>`_

        Examples:

        >>> sm.reprogramming_to_attractor(A=1, B=0)
        >>> sm.reprogramming_to_attractor({"A": 1, "B": 0})
        """
        strategies = ReprogrammingStrategies()
        spec = PartialState(*spec, **kwspec)
        for i, attractor in enumerate(self.attractors):
            if spec.match_state(attractor):
                for cs in self.control_sets[i]:
                    p = TemporaryPerturbation(cs)
                    s = FromCondition("input", p) if self.fixed else FromAny(p)
                    strategies.add(s)
        if self.fixed:
            strategies.register_alias("input", self.fixed)
        return strategies




from colomoto_jupyter import IN_IPYTHON, jupyter_setup

if IN_IPYTHON:
    jupyter_setup("stablemotifs")

    from colomoto_jupyter.upload import jupyter_upload
    def upload():
        return jupyter_upload("upload", "load")

else:
    def upload():
        raise NotImplementedError





from setuptools import setup, find_packages

NAME = 'stablemotifs'

setup(name=NAME,
    version="1.0",
    author = "Loïc Paulevé",
    author_email = "loic.pauleve@labri.fr",
    url = "https://github.com/algorecell/pyStableMotifs",
    description = "Python interface to StableMotifs",
    install_requires = [
        "colomoto_jupyter",
        "algorecell_types",
        "pandas",
    ],
    classifiers=[
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    keywords="computational systems biology",

    include_package_data = True,
    packages = find_packages(exclude=["examples"]),
    py_modules = ["stablemotifs_setup"]
)

